
/* ipqd.c - Daemon to wait for the host*/


/* Copyright (c) 2006, TeamF1 Inc. */


/*
modification history
--------------------
01a,24Feb06,srk created

*/



/*
DESCRIPTION

This module implements daemon program to wait for the hosts which
tries to connect the internet. This daemon compares the 
pattern and engine version with the Server TrendMicro System.

*/

/* includes */
#include <stdio.h> 
#include <stdlib.h> 
#include <errno.h> 
#include <string.h> 
#include <sys/types.h> 
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <sys/wait.h>
#include <pthread.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <libipq/libipq.h>

/* defines */

#define IPQ_DEBUG
#ifdef IPQ_DEBUG
#define ipqprintf(x...)  do {fprintf(stdout, x); fflush(stdout);}while(0)
#else
#define ipqprintf(x...)
#endif /* IPQ_DEBUG */


#define IPQD_PID_FILE     "/var/run/ipqd.pid"


#define GPL_RET_ERROR(errCode, errLabel) \
     do {                                                                   \
	     ipqprintf("At %s:%d -> %s\n", __FILE__, __LINE__,#errCode); \
            goto errLabel;                                                     \
        } while(0)


#define IPQ_RET_ERROR(errcode)   GPL_RET_ERROR(errCode, tmRetError)


#define TMGPL_CMD_BUF_SIZE       (128)
#define BUFSIZE         (1024)
#define SPACE   ' '     /* the space character */

#define CMDSTATUS 1

#define tmgpldCmdCat(cmdBuf, fmt...)                               \
        do {                                                     \
            result = snprintf(cmdBuf+bufOffSet,                  \
                              TMGPL_CMD_BUF_SIZE-bufOffSet, fmt); \
            if (result == 0)                                     \
                IPQ_RET_ERROR(TM_ERR_CMD_TOO_LONG);           \
            bufOffSet += result;                                 \
        } while(0)

/* typedef */


char ** tmgpldStrBreak ( char *);
int tmgpldCmdExec ( char *);
static void ipqdPid(void);

/*******************************************************************************
*
* tmgpldStrBreak - tokenize a string based on spaces into an argv[] array for main
*
* This routine breaks a string using spaces as delimiters between tokens and
* returns an array of strings suitable for tmgpldCmdExec(). The array must be
* freed later using free().
*
* RETURNS: pointer to array of strings or NULL
*/
char ** tmgpldStrBreak 
    (
    char * pString
    )
    {
    char ** ppArray = NULL;
    char * pPtr = pString;
    int tokenCount = 0;
    int i = 0;
        
    if (pString == NULL)
        IPQ_RET_ERROR(TM_ERR_INVALID_ARGUMENT);

    /* count number of tokens separated by spaces */
    while (1)
        {
        char c = *pPtr;
        char prev = *(pPtr-1);
        if (c == SPACE && prev != SPACE)
            tokenCount++;
        if (c == 0)
            {
            tokenCount++;
            break;
            }
        pPtr++;
        }

    if (tokenCount == 0)
        IPQ_RET_ERROR(TM_ERR_NO_TOKENS);

    /* 
     * allocate an array of pointers. 
     * number of pointers = number of tokens + termination null pointer 
     */
    ppArray = malloc((tokenCount + 1) * sizeof(char *));
    if (ppArray == NULL)
        IPQ_RET_ERROR(TM_ERR_NO_MEM);

    memset(ppArray, 0, (tokenCount + 1) * sizeof(char *));

    /* allocate memory for each pointer and copy words skipping all spaces */
    pPtr = pString;
    while (i < tokenCount)
        {
        char * pStart = NULL;
        char * pEnd = NULL;
        while (1)
            {
            if (pStart == NULL && (*pPtr) != SPACE)
                {
                pStart = pPtr;
                }
            if ((*pPtr == SPACE && pStart != NULL) || *pPtr == 0)
                {
                /* + 1 for null-termination */
                pEnd = pPtr + 1;
                break;
                }
            pPtr++;
            }
        
        ppArray[i] = malloc(pEnd - pStart);
        if (ppArray[i] == NULL)
            IPQ_RET_ERROR(ADPCMD_ERR_NO_MEM);
        memcpy(ppArray[i], pStart, pEnd - pStart);
        ppArray[i][pEnd - pStart - 1] = 0;
        i++;
        }

    return ppArray;

    tmRetError:
        {
        if (ppArray != NULL)
            {
            int i = tokenCount;
            while (i--)
                 free(ppArray[i]);
            }
        return NULL;
        }
    }


/**************************************/

int tmgpldCmdExec
    (
    char * pCmdString   /* command string */
    )
    {
    char ** ppArgArray = NULL;
    int cmdStatus = 0;
    int result = 0;
    pid_t pid;
    int fd;

    ipqprintf("tmclidCmdExec: exec'ing -> %s\n", pCmdString);
    ppArgArray = tmgpldStrBreak(pCmdString);
    if (ppArgArray == NULL)
        IPQ_RET_ERROR(TM_ERR_NESTED);
    pid = fork();
    switch (pid)
        {
        case -1:
            {
            IPQ_RET_ERROR(TM_ERR_SYSTEM | errno);
            break;
            }
        case 0:
            {
	    /* close all descriptors except 0,1,2 */
            for (fd = 3; fd <= 1024; fd++)
                close(fd);
	    
            result = execvp(ppArgArray[0], ppArgArray);
            if (result < 0)
                IPQ_RET_ERROR(TM_ERR_SYSTEM | errno);
            /* exec doesnt return and this is never reached*/
            }
        default:
            {
            /* parent will wait for child to exit */
            result = wait(&cmdStatus);
            ipqprintf("tmdCmdExec: %s -> cmdStatus: %d\n", 
                        pCmdString, cmdStatus);
            if (WIFEXITED(cmdStatus))
                return WEXITSTATUS(cmdStatus);
            else 
                return 0;
            }
        }


    tmRetError:
        {

	if(ppArgArray)
        free(ppArgArray);
        return 0;
        }
    }

/**********************************************************
*  ipqd Daemon 
*
*
*
*
*/

int main
    (
    int argc,
    char **ppArgv      
    )
    {
    int  cli_comm_port = 0;
    char pattern[10];
    char engine[10];


    int result;
    char cmdBuf[TMGPL_CMD_BUF_SIZE];
    int bufOffSet = 0;
    char * pCli_ip;
    char cli_ip;


    int status = 0;
    unsigned char buf[BUFSIZE];
    struct ipq_handle *h;
    struct iphdr *head;
    struct in_addr temp_ip;
    char *tmp;

        {
        int log = open("/var/log/ipq", O_CREAT | O_APPEND | O_RDWR, 0);
        dup2(log, STDOUT_FILENO);
        dup2(log, STDERR_FILENO);
        }
    if (argc < 1)
        IPQ_RET_ERROR(TM_INVALID_NUM_ARG);

    h = ipq_create_handle(0, PF_INET);
    if (!h)
        IPQ_RET_ERROR(TM_UNABLE_TO_CREATE_IPQ_HLD);

    status = ipq_set_mode(h, IPQ_COPY_PACKET, BUFSIZE);
    if (status < 0)
        IPQ_RET_ERROR(TM_UNABLE_TO_CREATE_IPQ_MODE);

    ipqprintf("ipqd started\n");	

    /*Set PID */
    ipqdPid();

    do
        {
        memset(buf,0,sizeof(buf));	

        status = ipq_read(h, buf, BUFSIZE, 0);
        if (status < 0)
            continue;

        switch (ipq_message_type(buf)) 
            {
            case NLMSG_ERROR:
                break;

            case IPQM_PACKET: 
                {
                ipq_packet_msg_t *m = ipq_get_packet(buf);

                head = (struct iphdr *)m->payload;
                temp_ip.s_addr = head->saddr;
                pCli_ip = inet_ntoa(temp_ip);

                memset(cmdBuf,0,sizeof(cmdBuf));
                bufOffSet = 0;

                snprintf(cmdBuf,TMGPL_CMD_BUF_SIZE,
                         "usrMisc TMCliConnect %s %s %s %s",
                         ppArgv[1], ppArgv[2], ppArgv[3], pCli_ip);

                result = tmgpldCmdExec(cmdBuf);  

                if(result == 1)
                    {
                    ipqprintf("Valid IP: %s\n", pCli_ip);	
                    status = ipq_set_verdict(h, m->packet_id,
                                             NF_ACCEPT, 0, NULL);
                    }
                else
                    ipqprintf("Invalid IP: %s\n", pCli_ip);	
                break;
                }

            default:
                ipqprintf("Unknown message type!\n");
                break;
            }
        } while (1);

    ipq_destroy_handle(h);
    tmRetError:
        {
        ipq_destroy_handle(h);	    
        return 0;
        }
    }

/****************************************
*
*
*/

static void ipqdPid
    (
    void
    )
    {
    FILE *fp = fopen(IPQD_PID_FILE,"wb+");
    if (fp == NULL)
        return;
    fprintf(fp,"%d \n",getpid()); 
    fclose(fp);
    return;
    }
